(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Equational_Reasoning_Impl
imports
  Equational_Reasoning
  Trs_Impl
begin

abbreviation "shows_eqs \<equiv> shows_trs' shows shows ''equational system:'' '' = ''"
abbreviation "shows_eq \<equiv> shows_rule' shows shows '' = ''"
abbreviation shows_ln :: "nat \<Rightarrow> shows" where
  "shows_ln i \<equiv> shows_nl +@+ shows i +@+ shows_string '': ''"

function
  eq_proof_lines :: "('f::show, 'v::show) eq_proof \<Rightarrow> nat \<Rightarrow> (shows \<times> nat \<times> ('f, 'v) rule)"
and eq_proofs_lines :: "('f::show, 'v::show) eq_proof list \<Rightarrow> nat \<Rightarrow> (shows \<times> nat list \<times> ('f, 'v) term list \<times> ('f, 'v) term list)"
where
  "eq_proof_lines (Refl s) i = (shows_ln (Suc i) +@+ shows_eq (s, s) +@+ shows_string '' [refl]'', Suc i, s, s)"
| "eq_proof_lines (Sym p) i = (
    let (s, i', l, r) = eq_proof_lines p i in
    (s +@+ shows_ln (Suc i') +@+ shows_eq (r, l) +@+ '' [sym '' +#+ shows i' +@+ shows_string '']'', Suc i', r, l))"
| "eq_proof_lines (Trans p1 p2) i = (
    let (s1, i1, s, t) = eq_proof_lines p1 i in
    let (s2, i2, u, v) = eq_proof_lines p2 i1 in
    (s1 +@+ s2 +@+
      shows_ln (Suc i2) +@+ shows_eq (s, v) +@+ '' [trans '' +#+ shows i1 +@+ '', '' +#+ shows i2 +@+ shows_string '']'', Suc i2, s, v))"
| "eq_proof_lines (Assm (l, r) \<sigma>) i = (
    let eq = (l \<cdot> \<sigma>, r \<cdot> \<sigma>) in
    (shows_ln (Suc i) +@+ shows_eq eq +@+ '' [assm '' +#+ shows_eq (l, r) +@+ shows_string '']'', Suc i, eq))"
| "eq_proof_lines (Cong f ps) i = (
    let (s, is, ls, rs) = eq_proofs_lines ps i in
    let eq = (Fun f ls, Fun f rs) in
    let i' = last is in
    let is = butlast is in
    (s +@+ shows_ln (Suc i') +@+ shows_eq eq +@+ shows_list_gen shows '' [cong]'' '' [cong '' '', '' '']'' is, Suc i', eq))"

| "eq_proofs_lines [] i = (id, [i], [], [])"
| "eq_proofs_lines (p#ps) i = (
    let (s1, i', l, r) = eq_proof_lines p i in
    let (s2, is, ls, rs) = eq_proofs_lines ps i' in
    (s1 +@+ s2, i'#is, l#ls, r#rs))"
  by (pat_completeness) auto
  termination by (lexicographic_order)

abbreviation shows_eq_proof where "shows_eq_proof p \<equiv> fst (eq_proof_lines p 0)"

definition
  check_estep ::
    "('f::show, 'v::show) equation list \<Rightarrow>
    pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow>
    shows check"
where
  "check_estep E p rule l_to_r s t \<equiv> do {
    check (\<exists>r\<in>set E. rule =v r)
      (shows_rule rule +@+ '' is not an equation of'' +#+ shows_nl +@+ shows_eqs E +@+ shows_nl);
    check (in_poss p s) (shows p +@+ '' is not a position of '' +#+ shows s +@+ shows_nl);
    check (in_poss p t) (shows p +@+ '' is not a position of '' +#+ shows t +@+ shows_nl);
    let C = ctxt_of_pos_term p s;
    let D = ctxt_of_pos_term p t;
    let u = subt_at s p;
    let v = subt_at t p;
    let rrule = (if l_to_r then rule else (snd rule, fst rule));
    let err = (''the term '' +#+ shows t
      +@+ '' does not result from a proper application of term '' +#+ shows s +@+ '' using equation ''
      +#+ shows_nl +@+ shows_eq rrule +@+ '' at position '' +#+ shows p +@+ shows_nl);
    (case match_list u [(u, fst rrule), (v, snd rrule)] of
      Some \<sigma> \<Rightarrow> check (C = D) err
    | None \<Rightarrow> error err)
  }"

lemma check_estep_sound:
  assumes ok: "isOK (check_estep E p r l_to_r s t)"
  shows "(s, t) \<in> estep (set E)"
proof -
  let ?C = "ctxt_of_pos_term p s"
  let ?D = "ctxt_of_pos_term p t"
  let ?r = "if l_to_r then r else (snd r,fst r)"
  let ?mlist = "[(subt_at s p, fst ?r), (subt_at t p, snd ?r)]"
  note ok = ok[unfolded check_estep_def Let_def, simplified]
  from ok
  have ps: "p \<in> poss s" and pt: "p \<in> poss t" and "\<exists>rule \<in> set E. r =v rule" by auto    
  then obtain rule where rule: "rule \<in> set E" and inst: "r =v rule" by auto
  from inst obtain \<delta> where fst: "fst r = fst rule \<cdot> \<delta>" and snd: "snd r = snd rule \<cdot> \<delta>"
    unfolding eq_rule_mod_vars_def instance_rule_def by auto
  let ?E = "(set E)\<^sup>\<leftrightarrow>"
  have "\<exists> rule \<in> ?E. fst ?r = fst rule \<cdot> \<delta> \<and> snd ?r = snd rule \<cdot> \<delta>"
  proof (cases l_to_r)
    case True
    thus ?thesis using rule fst snd by auto
  next
    case False
    show ?thesis
      by (rule bexI[of _ "(snd rule,fst rule)"], insert rule fst snd False, auto)
  qed
  then obtain rule where rule: "rule \<in> ?E" and fst: "fst ?r = fst rule \<cdot> \<delta>" and snd: "snd ?r = snd rule \<cdot> \<delta>" by auto
  from ok
    obtain \<sigma> where some: "match_list (subt_at s p) [(subt_at s p, fst ?r), (subt_at t p, snd ?r)] = Some \<sigma>"
    by force
  from match_list_sound[OF this] have "fst ?r \<cdot> \<sigma> = s |_ p" and "snd ?r \<cdot> \<sigma> = t |_ p" by auto
  hence s: "s |_ p = fst rule \<cdot> \<delta> \<circ>s \<sigma>" and t: "t = ?D\<langle>snd rule \<cdot> \<delta> \<circ>s \<sigma>\<rangle>"
    using ctxt_supt_id[OF pt] fst snd by simp+
  from ok and some
    have DC: "?D = ?C" by simp
  have "(s,t) \<in> rstep_r_p_s ?E rule p (\<delta> \<circ>s \<sigma>)"
    unfolding rstep_r_p_s_def Let_def
    by (rule, unfold split t[unfolded DC], intro conjI, rule ps, rule rule,
    insert ctxt_supt_id[OF ps, unfolded s], auto)
  hence "(s,t) \<in> rstep ?E" unfolding rstep_iff_rstep_r_p_s 
    using surjective_pairing by metis
  thus ?thesis unfolding estep_def .
qed

type_synonym  ('f, 'v) eseq = "(pos \<times> ('f, 'v) equation \<times> bool \<times> ('f, 'v) term) list"

fun
  check_conversion ::
    "('f::show, 'v::show) equation list \<Rightarrow> 
     ('f, 'v) eseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_conversion E [] s u = check (s = u) (
    ''the last term of the conversion '' +#+ shows_nl +@+ shows s +@+ shows_nl +@+
    ''does not correspond to the goal term''  +#+ shows_nl +@+ shows u +@+ shows_nl)"
| "check_conversion E ((p, r, l_to_r, t) # c) s u = do {
    check_estep E p r l_to_r s t;
    check_conversion E c t u
  }"

lemma check_conversion_sound:
  assumes ok: "isOK (check_conversion E eseq s t)"
  shows "(s, t) \<in> (estep (set E))^*"
  using ok
proof (induct eseq arbitrary: s t)
  case Nil
  thus ?case by auto
next
  case (Cons st eseq s u)
  obtain p r l_to_r t where st: "st = (p,r,l_to_r,t)" by (cases st, force+)
  from Cons(2)[unfolded st, simplified]
  have one: "isOK(check_estep E p r l_to_r s t)" and rec: "isOK(check_conversion E eseq t u)" by auto
  from check_estep_sound[OF one] Cons(1)[OF rec] show ?case by auto
qed

type_synonym ('f, 'v) subsumption_proof = "(('f, 'v) equation \<times> ('f, 'v) eseq) list"

definition
  check_subsumption_guided :: "('f::show, 'v::show) equation list \<Rightarrow> ('f, 'v) equation list \<Rightarrow> ('f, 'v) subsumption_proof \<Rightarrow> shows check"
where
  "check_subsumption_guided E E' convs \<equiv> do {
    check_subseteq E (map fst convs)
      <+? (\<lambda>r. ''could not find conversion for equation '' +#+ shows_eq r);
    check_allm (\<lambda>(e, seq). check_conversion E' seq (fst e) (snd e)
      <+? (\<lambda> s. ''problem in conversion for equation '' +#+ shows_eq e +@+ shows_nl +@+ s)) convs
  }"

lemma check_subsumption_guided:
  assumes ok: "isOK (check_subsumption_guided E E' convs)"
  shows "subsumes (set E') (set E)"
proof -
  note ok = ok[unfolded check_subsumption_guided_def, simplified]
  show ?thesis  
    unfolding subsumes_via_rule_conversion
  proof (clarify)
    fix s t
    assume "(s, t) \<in> set E"
    with ok obtain seq where conv: "((s, t), seq) \<in> set convs" by auto
    show "(s, t) \<in> (estep (set E'))^*"
      by (rule check_conversion_sound[of _ seq], insert conv ok, auto)
  qed
qed

lemma check_join_NF_conversion: 
  assumes ok: "isOK (check_join_NF R s t)"
  shows "(s,t) \<in> (estep (set R))^*"
  by (rule set_mp[OF join_conversion check_join_NF_sound[OF ok]])

definition
  check_subsumption_NF :: "('f::show, 'v::show) equation list \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_subsumption_NF E R \<equiv>  do {
    check_allm (\<lambda> e. check_join_NF R (fst e) (snd e)
      <+? (\<lambda>s. ''could not join equation '' +#+ shows_eq e +@+ shows_nl +@+ s)) E
  }"

lemma check_subsumption_NF:
  assumes ok: "isOK (check_subsumption_NF E R)"
  shows "subsumes (set R) (set E)"
proof -
  note ok = ok[unfolded check_subsumption_NF_def, simplified]
  show ?thesis
    unfolding subsumes_via_rule_conversion
  proof (clarify)
    fix s t
    assume "(s,t) \<in> set E"
    with ok have "isOK (check_join_NF R s t)" by auto
    from check_join_NF_conversion[OF this]
      show "(s, t) \<in> (estep (set R))^*" .
  qed
qed

definition
  check_subsumption :: "('f::show, 'v::show) equation list \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) subsumption_proof option \<Rightarrow> shows check"
where
  "check_subsumption E R convs_o \<equiv>
  (case convs_o of
    None \<Rightarrow> check_subsumption_NF E R
  | Some convs \<Rightarrow> check_subsumption_guided E R convs)"

lemma check_subsumption:
  assumes ok: "isOK (check_subsumption E R convs_o)"
  shows "subsumes (set R) (set E)"
proof -
  note ok = ok[unfolded check_subsumption_def]
  show ?thesis
  proof (cases convs_o)
    case None
    show ?thesis 
      by (rule check_subsumption_NF, insert None ok, auto)
  next
    case (Some convs)
    show ?thesis 
      by (rule check_subsumption_guided, insert Some ok, auto)
  qed
qed

end