(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Innermost_Usable_Rules_Impl
imports
  Innermost_Usable_Rules 
  Icap_Impl
  QDP_Framework_Impl
begin

fun is_ur_closed_term_impl 
  where "is_ur_closed_term_impl NFQ e_cap R U S (Var x) = True"
     |  "is_ur_closed_term_impl NFQ e_cap R U S (Fun f ts) = 
              ((\<forall> t \<in> set ts. is_ur_closed_term_impl NFQ e_cap R U S t) \<and>                
              (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> (case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> True | Some \<mu> \<Rightarrow> \<not> (\<forall> u \<in> set (args l). NFQ (mv_yvar u \<cdot> \<mu>)) \<or> \<not> (\<forall> u \<in> set S. NFQ (u \<cdot> \<mu>)))))"

(* TODO: clarify with Alex why one cannot use n within the zip-call (then the function package does not terminate) *)
declare size_simps[simp]
fun is_ur_closed_term_af_impl 
  where "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Var x) = True"
     |  "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Fun f ts) = (let n = length ts; \<pi>f = \<pi> (f,n) in
              ((\<forall> (i,t) \<in> set (zip [0 ..< length ts] ts). \<pi>f i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t) \<and>                
              (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> (case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> True | Some \<mu> \<Rightarrow> \<not> (\<forall> u \<in> set (args l). NFQ (mv_yvar u \<cdot> \<mu>)) \<or> \<not> (\<forall> u \<in> set S. NFQ (u \<cdot> \<mu>))))))"
declare size_simps[simp del]

declare is_ur_closed_term_af_impl.simps[code del]
lemma [code]: "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Fun f ts) = (let n = length ts; \<pi>f = \<pi> (f,n) in
              ((\<forall> (i,t) \<in> set (zip [0 ..< n] ts). \<pi>f i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t) \<and>                
              (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> (case mgu_class (Fun f (map e_cap ts)) l of
                 None \<Rightarrow> True | Some \<mu> \<Rightarrow> \<not> (\<forall> u \<in> set (args l). NFQ (mv_yvar u \<cdot> \<mu>)) \<or> \<not> (\<forall> u \<in> set S. NFQ (u \<cdot> \<mu>))))))"
               "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (Var x) = True"
  unfolding is_ur_closed_term_af_impl.simps Let_def by auto

lemma is_ur_closed_term_impl: fixes R :: "('f :: linorder,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set S)"
  shows "is_ur_closed_term_impl NFQ e_cap R U S = is_ur_closed_term' (set R) (set U) (set Q) icap full_af (set S)"
proof -
  let ?is_ur_closed_term = "is_ur_closed_term' (set R) (set U) (set Q) icap full_af"
  {
    fix t
    have "is_ur_closed_term_impl NFQ e_cap R U S t = ?is_ur_closed_term (set S) t"
    proof (induct t)
      case (Fun f ts)
      hence ball: "Ball (set ts) (is_ur_closed_term_impl NFQ e_cap R U S) = 
         (\<forall>i<length ts.
        full_af (f, length ts) i \<longrightarrow> ?is_ur_closed_term (set S) (ts ! i))" unfolding full_af_def[of "(f,length ts)"] set_conv_nth[of ts] by auto
      have all: "\<And> p q. \<lbrakk> \<And> l r. p l r = q l r\<rbrakk> \<Longrightarrow> (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> p l r) = (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> q l r)" by auto
      have ecap: "icap (set R) (set Q) (set S) = e_cap" unfolding e_cap ..
      have conj: "\<And> x y z u. x = y \<Longrightarrow> z = u \<Longrightarrow> (x \<and> z) = (y \<and> u)" by auto
      {
        fix x f 
        have "(case x of None \<Rightarrow> True | Some y \<Rightarrow> f y) = (\<not> (\<exists> y. x = Some y \<and> \<not> f y))" by (cases x, auto)
      } note id = this
      show ?case 
        unfolding is_ur_closed_term_impl.simps is_ur_closed_term'.simps
        by (rule conj[OF ball all], unfold ecap NFQ id, simp) blast
    qed simp
  } note main = this
  show ?thesis
    by (rule ext, rule main)
qed

lemma is_ur_closed_term_af_impl: fixes R :: "('f :: linorder,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set S)"
  shows "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S = is_ur_closed_term' (set R) (set U) (set Q) icap \<pi> (set S)"
proof -
  let ?is_ur_closed_term = "is_ur_closed_term' (set R) (set U) (set Q) icap \<pi>"
  {
    fix t
    have "is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t = ?is_ur_closed_term (set S) t"
    proof (induct t)
      case (Fun f ts)
      hence ball: "(\<forall> i < length ts. \<pi> (f,length ts) i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S (ts ! i)) = 
         (\<forall>i<length ts.
        \<pi> (f, length ts) i \<longrightarrow> ?is_ur_closed_term (set S) (ts ! i))" (is "?l' = ?r")
        using Fun[unfolded set_conv_nth[of ts]] by auto
      have id: "?l' = (\<forall> (i,t) \<in> set (zip [0 ..< length ts] ts). \<pi> (f,length ts) i \<longrightarrow> is_ur_closed_term_af_impl NFQ e_cap \<pi> R U S t)" unfolding set_zip by force
      note ball = ball[unfolded id]
      have all: "\<And> p q. \<lbrakk> \<And> l r. p l r = q l r\<rbrakk> \<Longrightarrow> (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> p l r) = (\<forall> (l,r) \<in> set R. (l,r) \<in> set U \<or> q l r)" by auto
      have ecap: "icap (set R) (set Q) (set S) = e_cap" unfolding e_cap ..
      have conj: "\<And> x y z u. x = y \<Longrightarrow> z = u \<Longrightarrow> (x \<and> z) = (y \<and> u)" by auto
      {
        fix x f 
        have "(case x of None \<Rightarrow> True | Some y \<Rightarrow> f y) = (\<not> (\<exists> y. x = Some y \<and> \<not> f y))" by (cases x, auto)
      } note id = this
      show ?case 
        unfolding is_ur_closed_term_af_impl.simps is_ur_closed_term'.simps Let_def
        by (rule conj[OF ball all], unfold ecap NFQ id, simp) blast
    qed simp
  } note main = this
  show ?thesis
    by (rule ext, rule main)
qed

definition is_ur_closed_term_impl_mv 
  where "is_ur_closed_term_impl_mv NFQ e_cap R U S \<equiv> let urc = is_ur_closed_term_impl NFQ e_cap R U (map mv_xvar S) in (\<lambda> t. urc (mv_xvar t))"

definition is_ur_closed_term_af_impl_mv 
  where "is_ur_closed_term_af_impl_mv NFQ e_cap \<pi> R U S \<equiv> let urc = is_ur_closed_term_af_impl NFQ e_cap \<pi> R U (map mv_xvar S) in (\<lambda> t. urc (mv_xvar t))"

lemma is_ur_closed_term_impl_mv: fixes R :: "('f :: linorder,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set (map mv_xvar S))"
  shows "is_ur_closed_term_impl_mv NFQ e_cap R U S t = is_ur_closed_term_mv' (set R) (set U) (set Q) icap' full_af (set S) t" 
  unfolding
    is_ur_closed_term_impl_mv_def 
    is_ur_closed_term_impl[OF NFQ e_cap]
    set_map
    is_ur_closed_term_mv_icap' Let_def ..

lemma is_ur_closed_term_af_impl_mv: fixes R :: "('f :: linorder,string)rules" and Q :: "('f,string)term list"
  assumes NFQ: "NFQ = (\<lambda> t. t \<in> NF_terms (set Q))"
  and e_cap: "e_cap = icap (set R) (set Q) (set (map mv_xvar S))"
  shows "is_ur_closed_term_af_impl_mv NFQ e_cap \<pi> R U S t = is_ur_closed_term_mv' (set R) (set U) (set Q) icap' \<pi> (set S) t" 
  unfolding
    is_ur_closed_term_af_impl_mv_def 
    is_ur_closed_term_af_impl[OF NFQ e_cap]
    set_map
    is_ur_closed_term_mv_icap' Let_def ..

definition is_ur_closed_impl_dpp_mv   :: "('d, 'f :: {linorder,show}, string) dpp_ops \<Rightarrow> 'd
  \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)term list \<Rightarrow> ('f,string)term \<Rightarrow> bool"
  where "is_ur_closed_impl_dpp_mv I d \<equiv>
  let ic = icap_impl_dpp I d;
      qnf = dpp_ops.is_QNF I d;
      r = dpp_ops.rules I d;
      urc = (\<lambda> S. is_ur_closed_term_impl qnf (ic S) r)
    in (\<lambda> U S. let S' = map mv_xvar S in (\<lambda> t. urc S' U S' (mv_xvar t)))"

definition is_ur_closed_af_impl_dpp_mv   :: "('d, 'f :: {linorder,show}, string) dpp_ops \<Rightarrow> 'd \<Rightarrow> 'f af
  \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)term list \<Rightarrow> ('f,string)term \<Rightarrow> bool"
  where "is_ur_closed_af_impl_dpp_mv I d \<pi> \<equiv>
  let ic = icap_impl_dpp I d;
      qnf = dpp_ops.is_QNF I d;
      r = dpp_ops.rules I d;
      urc = (\<lambda> S. is_ur_closed_term_af_impl qnf (ic S) \<pi> r)
    in (\<lambda> U S. let S' = map mv_xvar S in (\<lambda> t. urc S' U S' (mv_xvar t)))"

lemma is_ur_closed_impl_dpp_mv:
  fixes I::"('d, 'f::{show,linorder}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and r: "r \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
  shows "is_ur_closed_impl_dpp_mv I d u s t = is_ur_closed_term_mv' (set r) (set u) (set q) icap' full_af (set s) t" 
proof -
  interpret dpp_spec I by fact
  show ?thesis
    unfolding is_ur_closed_term_impl_mv[OF refl refl, symmetric]
      is_ur_closed_impl_dpp_mv_def Let_def is_QNF_sound icap_impl_dpp_icap[OF I] r q 
    unfolding is_ur_closed_term_impl_mv_def Let_def ..
qed    

lemma is_ur_closed_af_impl_dpp_mv:
  fixes I::"('d, 'f::{show,linorder}, string) dpp_ops" and d::"'d"
  defines q: "q \<equiv> dpp_ops.Q I d"
      and r: "r \<equiv> dpp_ops.rules I d"
  assumes I: "dpp_spec I"
  shows "is_ur_closed_af_impl_dpp_mv I d \<pi> u s t = is_ur_closed_term_mv' (set r) (set u) (set q) icap' \<pi> (set s) t" 
proof -
  interpret dpp_spec I by fact
  show ?thesis
    unfolding is_ur_closed_term_af_impl_mv[OF refl refl, symmetric]
      is_ur_closed_af_impl_dpp_mv_def Let_def is_QNF_sound icap_impl_dpp_icap[OF I] r q 
    unfolding is_ur_closed_term_af_impl_mv_def Let_def ..
qed    

definition
  usable_rules_proc ::
    "('dpp, 'f :: {linorder,show}, string) dpp_ops \<Rightarrow> 
     ('f,string)rules \<Rightarrow> 'dpp proc"
where
  "usable_rules_proc I U dpp \<equiv> 
    check_return (do {
     check (dpp_ops.NFQ_subset_NF_rules I dpp) (shows ''innermost rewriting required'');
     let P = dpp_ops.pairs I dpp;
     let urc = is_ur_closed_impl_dpp_mv I dpp U;
     let check_urc = (\<lambda> S t. check (urc S t) (shows ''term '' +@+ shows t +@+ shows '' is not closed under usable rules''));
     check_allm (\<lambda>(l, r). do {
        check_subseteq (vars_term_list r) (vars_term_list l) <+? (\<lambda> x. shows ''variable condition in P violated'');
        check_urc [l] r
     }) P;       
     check_allm (\<lambda>(l,r). check_urc (args l) r) U
   })
   (dpp_ops.intersect_rules I dpp U)"


lemma usable_rules_proc: assumes I: "dpp_spec I"
 shows "dpp_spec.sound_proc_impl I (usable_rules_proc I (U :: ('f :: {linorder,show},string)rules))"
proof -
  from assms interpret dpp_spec I .
  show ?thesis
  proof
    fix d d'    
    assume ok: "usable_rules_proc I U d = Inr d'" and fin: "finite_dpp (dpp d')"
    let ?P = "set (P d)"
    let ?Pw = "set (Pw d)"
    let ?Q = "set (Q d)"
    let ?R = "set (R d)"
    let ?Rw = "set (Rw d)"
    let ?U = "set U"
    let ?Rb = "?R \<union> ?Rw"
    let ?Rb' = "set (rules d)" 
    note ok = ok[unfolded usable_rules_proc_def dpp_spec_sound Let_def
      is_ur_closed_impl_dpp_mv[OF I], simplified] 
    from ok have NF: "NF_terms ?Q \<subseteq> NF_trs ?Rb" and NF': "NF_terms ?Q \<subseteq> NF_trs ?Rb'" by auto
    from ok have d': "d' = intersect_rules d U" by auto
    have d': "dpp d' = (?P, ?Pw, ?Q, ?R \<inter> ?U, ?Rw \<inter> ?U)" 
      unfolding d' unfolding intersect_rules_sound ..
    note fin = fin[unfolded this]
    interpret R_Q_U_ecap ?Rb' ?U ?Q icap'
      by (unfold_locales, rule icap, rule NF')
    have Rb: "?Rb = ?Rb'" by simp
    show "finite_dpp (dpp d)" unfolding dpp_spec_sound
    proof (rule usable_rules_proc[OF fin _ _ _ _])
      show "?Rb \<subseteq> ?Rb'" by simp
    next
      fix l r
      assume "(l,r) \<in> ?P \<union> ?Pw" 
      with ok 
      show "vars_term r \<subseteq> vars_term l" by blast
    next
      fix l r
      assume "(l,r) \<in> ?P \<union> ?Pw" 
      with ok
      show "is_ur_closed_term_mv full_af {l} r" by auto
    next
      fix l r
      assume "(l,r) \<in> ?U"
      with ok
      show "is_ur_closed_term_mv full_af (set (args l)) r" by auto
    qed
  qed
qed 

end