(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Reduction_Pair
imports
  Term_Order
  QDP_Framework 
begin


context redtriple
begin
lemma rel_SN:   
  shows "rel_SN (S,NS)"
  unfolding rel_SN_def rel_step.simps
  by (rule SN_subset[OF compatible_SN'[OF NS.compat SN]], regexp)

lemma rel_SN_rstep:
  assumes mono: "one_step_ctxt_closed S"
  shows "rel_SN (rstep S, rstep NS)" 
  by (rule rel_SN_mono[OF trs_inclusion[OF mono subst_S subset_refl] trs_inclusion[OF wm_NS subst_NS subset_refl] rel_SN])

lemma mono_redpair_sound_ichain: 
  assumes mono: "one_step_ctxt_closed S"
  and R: "R \<subseteq> NS \<union> S"
  and P: "P \<subseteq> NS \<union> S"
  shows "\<not> ichain (S \<inter> P, P - S, Q, S \<inter> R, R - S) s t \<sigma>"
  by (rule rel_SN_ichain, rule rel_SN_mono[OF qrstep_mono[of _ S _ "{}"] qrstep_mono[of _ NS _ "{}"]], insert R P rel_SN_rstep[OF mono], auto)

lemma mono_redpair_sound_ideriv: 
  assumes mono: "one_step_ctxt_closed S"
  and R: "R \<subseteq> NS \<union> S"
  shows "\<not> qrs_ideriv Q (S \<inter> R) (R - S) ts"
  using rel_SN_rstep[OF mono] qrs_ideriv_mono[of "S \<inter> R" S "R - S" NS Q "{}", of ts]
    R unfolding qrs_ideriv_def rel_SN_ideriv by auto
end

fun mono_redpair_proc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)dpp \<Rightarrow> ('f,'v)dpp \<Rightarrow> bool"
 where "mono_redpair_proc S NS Ps Rs (P,Pw,Q,R,Rw) (P',Pw',Q',R',Rw') = (one_step_ctxt_closed S \<and> R \<union> Rw' \<subseteq> NS \<union> S \<and> P \<union> Pw \<subseteq> NS \<union> S \<and> (P \<union> Pw) \<inter> Ps \<subseteq> S \<and> (R \<union> Rw) \<inter> Rs \<subseteq> S \<and> P' = P - Ps \<and> Pw' = Pw - Ps \<and> R' = R - Rs \<and> Rw' = Rw - Rs \<and> Q' = Q)"

fun mono_redpair_tt :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)qreltrs \<Rightarrow> ('f,'v)qreltrs \<Rightarrow> bool"
 where "mono_redpair_tt S NS Rs (Q,R,Rw) (Q',R',Rw') = (one_step_ctxt_closed S \<and> R \<union> Rw' \<subseteq> NS \<union> S \<and> (R \<union> Rw) \<inter> Rs \<subseteq> S \<and> R' = R - Rs \<and> Rw' = Rw - Rs \<and> Q' = Q)"

context redtriple
begin
lemma mono_redpair_proc_ci: "ci_proc (mono_redpair_proc S NS Ps Rs)"
unfolding ci_proc_def
proof (intro allI impI)
  fix P Pw Q R Rw P' Pw' Q' R' Rw'
  assume proc: "mono_redpair_proc S NS Ps Rs (P,Pw,Q,R,Rw) (P',Pw',Q',R',Rw')"  
  from proc have mono: "one_step_ctxt_closed S" and  R: "R' = R - Rs" and Rw: "Rw' = Rw - Rs" and Q: "Q' = Q" and P: "P' = P - Ps"
    and Pw: "Pw' = Pw - Ps" by auto
  show "R' \<union> Rw' \<subseteq>ci R \<union> Rw \<and> Q = Q' \<and> (\<forall> s t \<sigma> . min_ichain (P, Pw, Q, R, Rw) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (P',Pw',Q',R', Rw') (shift s i) (shift t i) (shift \<sigma> i)))"  
  proof (intro conjI, unfold R Rw Q P Pw, rule ci_subsetI, blast, simp, intro allI impI)
    fix s t \<sigma>
    assume min: "min_ichain (P,Pw,Q,R,Rw) s t \<sigma>"
    hence ichain: "ichain (P,Pw,Q,R,Rw) s t \<sigma>" by simp
    show "\<exists> i. ichain (P - Ps, Pw - Ps, Q, R - Rs, Rw - Rs) (shift s i) (shift t i) (shift \<sigma> i)"
    proof (rule ichain_split[OF ichain], rule)
      assume chain: "ichain (Ps \<inter> (P \<union> Pw), P \<union> Pw - Ps ,Q,Rs \<inter> (R \<union> Rw) ,R \<union> Rw - Rs) s t \<sigma>"
      have "ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S ,Q, S \<inter> (R \<union> Rw), R \<union> Rw - S) s t \<sigma>"
        by (rule ichain_mono[OF chain], insert proc, auto)
      moreover have "\<not> ichain (S \<inter> (P \<union> Pw), P \<union> Pw - S ,Q, S \<inter> (R \<union> Rw),R \<union> Rw - S) s t \<sigma>"
        by (rule mono_redpair_sound_ichain[OF mono], insert proc, auto)
      ultimately show False by simp
    qed
  qed
qed

lemma mono_redpair_tt: "rel_tt (mono_redpair_tt S NS Rs)"
unfolding rel_tt_def
proof (intro allI impI)
  fix Q R Rw Q' R' Rw' ts
  assume proc: "mono_redpair_tt S NS Rs (Q,R,Rw) (Q',R',Rw')" 
    and ideriv: "qrs_ideriv Q R Rw ts"
  from proc have mono: "one_step_ctxt_closed S" and  R: "R' = R - Rs" and Rw: "Rw' = Rw - Rs" and Q: "Q' = Q" and Rns: "R \<union> Rw \<subseteq> NS \<union> S" by auto
  from qrs_ideriv_split[OF ideriv mono_redpair_sound_ideriv[OF mono Rns]]
  obtain i where ideriv: "qrs_ideriv Q (R - S) (Rw - S) (shift ts i)" ..
  have "qrs_ideriv Q (R - Rs) (Rw - Rs) (shift ts i)"
    by (rule qrs_ideriv_mono[OF _ _ _ ideriv], insert proc, auto)
  thus "Ex (qrs_ideriv Q' R' Rw')" unfolding R Rw Q by auto
qed

lemma manna_ness:
  assumes mono: "one_step_ctxt_closed S"
  and strict: "R \<subseteq> S"
  shows "SN (rstep R)"
proof -
  from mono strict subst_S trs_inclusion have "(rstep R) \<subseteq> S" by auto
  from SN_subset[OF SN this] show "SN (rstep R)"  .
qed
end

end